"""Database connection helpers."""
import os
from typing import Tuple


def db_uri_with_origin() -> Tuple[str, str]:
    """Determine the default db uri.

    Returns:
        Tuple[DB URI, the origin of the URI]
    """
    uri = os.environ.get("POKEDEX_DB_ENGINE", None)
    origin = "environment"

    if uri is None:
        import importlib_resources

        sqlite_path = importlib_resources.files("pokemaster2").parent / "data/pokedex.sqlite3"
        uri = "sqlite:///" + str(sqlite_path)
        origin = "default"

    return uri, origin


def csv_dir_with_origin() -> Tuple[str, str]:
    """Determine the default csv path.

    Returns:
        Tuple[CSV file path, the origin of the CSV path]
    """
    csv_path = os.environ.get("POKEDEX_CSV_PATH", None)
    origin = "environment"

    if csv_path is None:
        import importlib_resources

        csv_path = str(importlib_resources.files("pokemaster2").parent / "data/csv")
        origin = "default"

    return csv_path, origin


def db_uri() -> str:
    """Get only the default db uri, omitting the origin."""
    return db_uri_with_origin()[0]


def csv_dir() -> str:
    """Get only the default csv path, omitting the origin."""
    return csv_dir_with_origin()[0]



"""Base Pokemon."""
import operator
from typing import Callable, Sequence, Type, TypeVar, Union

import attr

from pokemaster2.prng import PRNG

S = TypeVar("S", bound="Stats")
P = TypeVar("P", bound="BasePokemon")


STAT_NAMES = ["hp", "atk", "def_", "spatk", "spdef", "spd"]
STAT_NAMES_FULL = {
    "hp": "hp",
    "attack": "atk",
    "defense": "def_",
    "special-attack": "spatk",
    "special-defense": "spdef",
    "speed": "spd",
}

prng = PRNG()


@attr.s(auto_attribs=True)
class Stats:
    """Generic stats, can be used for Pokemon stats/IV/EV."""

    hp: int
    atk: int
    def_: int
    spatk: int
    spdef: int
    spd: int

    def __add__(self: S, other: Union[S, int]) -> S:
        """Pointwise addition."""
        return self._make_operator(operator.add, other)

    def __sub__(self: S, other: Union[S, int]) -> S:
        """Pointwise subtraction."""
        return self._make_operator(operator.sub, other)

    def __mul__(self: S, other: Union[S, int]) -> S:
        """Pointwise multiplication."""
        return self._make_operator(operator.mul, other)

    def __floordiv__(self: S, other: Union[S, int]) -> S:
        """Pointwise floor division."""
        return self._make_operator(operator.floordiv, other)

    __radd__ = __add__
    __rmul__ = __mul__

    def _make_operator(
        self: S,
        operator: Callable[[int, int], int],
        other: Union[S, int],
    ) -> S:
        """Programmatically create point-wise operators.

        Args:
            operator: A callable (Real, Real) -> Real.
            other: If `other` is a `Stats` instance, then the
                operator will be applied point-wisely. If `other` is a
                number, then a scalar operation will be applied.

        Raises:
            TypeError: `other` should be either another `Stats` or `int`.

        Returns:
            A `Stats` instance.
        """
        names = (
            "hp",
            "atk",
            "def_",
            "spatk",
            "spdef",
            "spd",
        )

        if not isinstance(other, type(self)) and not isinstance(other, int):
            raise TypeError(
                f"unsupported operand type(s) for {operator}: "
                f"'{type(self)}' and '{type(other)}'"
            )
        result_stats = {}
        for stat in names:
            if isinstance(other, type(self)):
                result_stats[stat] = int(operator(getattr(self, stat), getattr(other, stat)))
            elif isinstance(other, int):
                result_stats[stat] = int(operator(getattr(self, stat), other))
        return self.__class__(**result_stats)

    def validate_iv(self: S) -> bool:
        """Check if each IV is between 0 and 32."""
        for stat in STAT_NAMES:
            if not 0 <= getattr(self, stat) <= 32:
                raise ValueError(
                    f"The {stat} IV ({getattr(self, stat)}) must be a number "
                    "between 0 and 32 inclusive."
                )
        return True

    @classmethod
    def create_iv(cls: Type[S], gene: int) -> S:
        """Create IV stats from a Pokémon's gene.

        Args:
            gene: An `int` generated by the PRNG.

        Returns:
            A `Stats` instance.
        """
        return cls(
            hp=gene % 32,
            atk=(gene >> 5) % 32,
            def_=(gene >> 10) % 32,
            spd=(gene >> 16) % 32,
            spatk=(gene >> 21) % 32,
            spdef=(gene >> 26) % 32,
        )

    @classmethod
    def zeros(cls: Type[S]) -> S:
        """Empty Stats."""
        return cls(
            hp=0,
            atk=0,
            def_=0,
            spatk=0,
            spdef=0,
            spd=0,
        )

    @classmethod
    def nature_modifiers(cls: Type[S], nature: str) -> S:
        """Generate nature modifier Stats."""
        # nature_data = _db.get_nature(identifier=nature)
        # modifiers = {}
        # for stat in STAT_NAMES:
        #     modifiers[stat] = 1
        # if nature_data.is_neutral:
        #     return cls(**modifiers)
        # modifiers[STAT_NAMES_FULL[nature_data.increased_stat.identifier]] = 1.1
        # modifiers[STAT_NAMES_FULL[nature_data.decreased_stat.identifier]] = 0.9
        # return cls(**modifiers)


@attr.s(auto_attribs=True)
class BasePokemon:
    """The underlying structure of a Pokémon.

    No fancy initializations, no consistency checks, just a very basic
    Pokémon model. Anything is possible with this BasePokemon. This
    class also contains common and basic behaviors of Pokémon, such as
    leveling-up, learning/forgetting moves, evolving into another
    Pokémon, etc.

    This class is never meant to be instantiated directly.
    """

    national_id: int
    species: str
    types: Sequence[str]
    item_held: str
    exp: int
    level: int

    base_stats: Stats
    iv: Stats
    current_stats: Stats
    stats: Stats
    ev: Stats

    # move_set = Mapping[int, Mapping[str, Union[str, int]]]
    pid: str
    gender: str
    nature: str
    ability: str

    # def evolve(self: P) -> None:
    #     """
    #     Evolve into another Pokémon.

    #     1. Statistics are updated.
    #     2. Learnset is updated.
    #     3. Evolution tree is updated.

    #     Returns:
    #         Nothing

    #     """
    #     pass

    # def level_up(self: P) -> None:
    #     """Increase `Pokemon`'s level by 1.

    #     Returns:
    #         Nothing
    #     """
    #     pass

    # @classmethod
    # def _from_pokedex_by_id(
    #     cls: "BasePokemon",
    #     national_id: int,
    #     level: int,
    #     item_held: str = None,
    #     iv: Stats = None,
    #     ev: Stats = None,
    #     pid: int = None,
    #     nature: str = None,
    #     ability: str = None,
    #     gender: str = None,
    # ) -> "BasePokemon":
    #     """Instantiate a `BasePokemon` by its national id.

    #     Everything else is randomized.

    #     Args:
    #         national_id: the Pokemon's ID in the National Pokedex.
    #         level: Pokemon's level.
    #         item_held: Pokemon's holding item.
    #         iv: Pokemon's individual values, `Stats`, used to determine its permanent stats.
    #             A random IV will be set if not provided.
    #         ev: Pokemon's effort values, `Stats`, used to determine its permanent stats. An
    #             all-zero ev will be set if not provided.
    #         pid: Pokemon's personality id. `nature`, `ability`, and `gender` will use
    #             their provided value first. A random `pid` will be set if not provided.
    #         nature: Pokemon's nature, used to determine its permanent stats. If nothing is
    #             provided, then the function will use `pid` to determine its `nature`.
    #         ability: Pokemon's ability, `str`. If nothing is provided, then the function
    #             will use `pid` to determine its `nature`.
    #         gender: Pokemon's gender.  If nothing is provided, then the function will use
    #             `pid` to determine its `nature`.

    #     Returns:
    #         A `BasePokemon` instance.
    #     """
    # # Build pokemon data
    # pokemon_data = _db.get_pokemon(national_id=national_id)
    # growth_data = _db.get_experience(national_id=national_id, level=level)
    # species_data = pokemon_data.species
    # species = species_data.identifier

    # # Determine stats
    # gene = prng.create_gene()
    # iv = iv or Stats.create_iv(gene=gene)
    # ev = ev or Stats.zeros()
    # base_stats = {}
    # for i, stat in enumerate(STAT_NAMES):
    #     base_stats[stat] = pokemon_data.stats[i].base_stat
    # stats = _calc_stats(level=level, base_stats=base_stats, iv=iv, ev=ev, nature=nature)
    # current_stats = stats

    # # PID related attributes
    # pid = pid or prng.create_personality()
    # nature = nature or _db.get_nature(pid).identifier
    # ability = ability or _db.get_ability(species=species, personality=pid).identifier
    # gender = gender or _db.get_pokemon_gender(species=species, personality=pid).identifier

    # return cls(
    #     pid=pid,
    #     national_id=species_data.id,
    #     species=species,
    #     types=list(map(lambda x: x.identifier, pokemon_data.types)),
    #     item_held=item_held,
    #     exp=growth_data.experience,
    #     level=growth_data.level,
    #     stats=stats,
    #     current_stats=current_stats,
    #     ev=ev,
    #     iv=iv,
    #     nature=nature,
    #     ability=ability,
    #     gender=gender,
    # )


def _calc_stats(level: int, base_stats: Stats, iv: Stats, ev: Stats, nature: str) -> Stats:
    """Calculate the Pokemon's stats."""
    nature_modifiers = Stats.nature_modifiers(nature)
    residual_stats = Stats(
        hp=10 + level,
        atk=5,
        def_=5,
        spatk=5,
        spdef=5,
        spd=5,
    )

    stats = ((base_stats * 2 + iv + ev // 4) * level // 100 + residual_stats) * nature_modifiers
    if base_stats.hp == 1:
        stats.hp = 1
    return stats



"""The pokedex database models."""
from typing import List

import peewee

database = peewee.SqliteDatabase(None)


class BaseModel(peewee.Model):
    """BaseModel for all pokdex Models."""

    class Meta:
        """All BaseModel uses database."""

        database = database


class Pokemon(BaseModel):
    """A Pokémon.  The core to this whole mess.

    This table defines "Pokémon" the same way the games do: a form with
    different types, moves, or other game-changing properties counts as a
    different Pokémon.  For example, this table contains four rows for Deoxys,
    but only one for Unown.

    Non-default forms have IDs above 10000.
    IDs below 10000 match the species_id column, for convenience.
    """

    id = peewee.IntegerField(primary_key=True)  # noqa: A003
    identifier = peewee.CharField(
        max_length=79,
        help_text="An identifier, including form iff this row corresponds to a single, named form",
    )
    species_id = peewee.IntegerField(
        help_text="ID of the species this Pokémon belongs to",
    )
    height = peewee.IntegerField(
        help_text="The height of the Pokémon, in tenths of a meter (decimeters)",
    )
    weight = peewee.IntegerField(
        help_text="The weight of the Pokémon, in tenths of a kilogram (hectograms)",
    )
    base_experience = peewee.IntegerField(
        help_text="The base EXP gained when defeating this Pokémon",
    )
    order = peewee.IntegerField(
        index=True,
        help_text=(
            "Order for sorting. Almost national order, " "except families are grouped together."
        ),
    )
    is_default = peewee.BooleanField(
        index=True,
        help_text="Set for exactly one pokemon used as the default for each species.",
    )


class PokemonSpecies(BaseModel):
    """A Pokémon species: the standard 1–151.  Or 649.  Whatever.

    ID matches the National Pokédex number of the species.
    """

    id = peewee.IntegerField(primary_key=True)  # noqa: A003
    identifier = peewee.CharField(
        max_length=79,
        help_text="An identifier",
    )
    # generation_id = peewee.ForeignKeyField(
    #     "Generation",
    #     backref="pokemon_species",
    #     help_text="ID of the generation this species first appeared in",
    # )
    evolves_from_species_id = peewee.ForeignKeyField(
        "self",
        backref="evolves_to",
        null=True,
        help_text="The species from which this one evolves",
    )
    # evolution_chain_id = peewee.ForeignKeyField(
    #     "EvolutionChain",
    #     backref="pokemon_species",
    #     help_text="ID of the species' evolution chain (a.k.a. family)",
    # )
    # color_id = peewee.ForeignKeyField(
    #     "PokemonColor",
    #     backref="pokemon_species",
    #     help_text="ID of this Pokémon's Pokédex color, as used for a gimmick search function in the games.",
    # )
    # shape_id = peewee.ForeignKeyField(
    #     "PokemonShape",
    #     backref="pokemon_species",
    #     help_text="ID of this Pokémon's body shape, as used for a gimmick search function in the games.",
    # )
    # habitat_id = peewee.ForeignKeyField(
    #     "PokemonHabitat",
    #     backref="pokemon_species",
    #     null=True,
    #     help_text="ID of this Pokémon's habitat, as used for a gimmick search function in the games.",
    # )
    gender_rate = peewee.IntegerField(
        help_text="The chance of this Pokémon being female, in eighths; or -1 for genderless",
    )
    capture_rate = peewee.IntegerField(help_text="The base capture rate; up to 255")
    base_happiness = peewee.IntegerField(help_text="The tameness when caught by a normal ball")
    is_baby = peewee.BooleanField(
        help_text="True iff the Pokémon is a baby, i.e. a lowest-stage Pokémon that cannot breed but whose evolved form can.",
    )
    hatch_counter = peewee.IntegerField(
        help_text="Initial hatch counter: one must walk 255 × (hatch_counter + 1) steps before this Pokémon's egg hatches, unless utilizing bonuses like Flame Body's",
    )
    has_gender_differences = peewee.BooleanField(
        help_text="Set iff the species exhibits enough sexual dimorphism to have separate sets of sprites in Gen IV and beyond.",
    )
    # growth_rate_id = peewee.ForeignKeyField(
    #     "PokemonGrowthRate",
    #     backref="pokemon_species",
    #     help_text="ID of the growth rate for this family",
    # )
    forms_switchable = peewee.BooleanField(
        help_text="True iff a particular individual of this species can switch between its different forms.",
    )
    order = peewee.IntegerField(
        index=True,
        help_text="The order in which species should be sorted.  Based on National Dex order, except families are grouped together and sorted by stage.",
    )
    conquest_order = peewee.IntegerField(
        null=True,
        index=True,
        help_text="The order in which species should be sorted for Pokémon Conquest-related tables.  Matches gallery order.",
    )


Pokemon.species = peewee.ForeignKeyField(PokemonSpecies)


def get_pokemon(identifier: str) -> List[Pokemon]:
    """Find a single `Pokemon` instance."""
    pokemon_set = (
        Pokemon.select(Pokemon, PokemonSpecies)
        .join(PokemonSpecies, on=(Pokemon.species_id == PokemonSpecies.id))
        .where(Pokemon.identifier == identifier)
    )

    return pokemon_set


MODELS = [Pokemon, PokemonSpecies]



"""Load csv files into database."""
import csv
from pathlib import Path
from typing import Optional, Sequence

import peewee
from loguru import logger

from pokemaster2.db import default, tables

# from playhouse import db_url


def get_database(uri: Optional[str] = None) -> peewee.SqliteDatabase:
    """Connect to and return a database."""
    if uri is None:
        uri, origin = default.db_uri_with_origin()
    # elif not uri.startswith("sqlite://")
    else:
        uri, origin = uri, "command-line"

    # database = db_url.connect(uri)
    database = peewee.SqliteDatabase(uri)
    if database.connect():
        logger.debug(
            "Connected to database {database} (from {origin}).", database=uri, origin=origin
        )
    else:
        logger.error(
            "Failed to connect to database {database} (from {origin}", database=uri, origin=origin
        )

    return database


def get_csv_dir(csv_dir: Optional[str] = None) -> str:
    """Return the csv dir we are about to use."""
    if csv_dir is None:
        csv_dir, origin = default.csv_dir_with_origin()
    else:
        csv_dir, origin = csv_dir, "command-line"

    logger.debug("Using csv directory {csv_dir} (from {origin}).", csv_dir=csv_dir, origin=origin)
    return csv_dir


def load(
    database: peewee.SqliteDatabase,
    csv_dir: str,
    models: Sequence[tables.BaseModel] = None,
    drop_tables: bool = False,
    safe: bool = True,
    recursive: bool = True,
    # langs: Optional[str] = None,
) -> None:
    """Load data from CSV files into the given database.

    Args:
        database: `peewee` database to use.
        models: List of tables to load. If omitted, all tables are loaded.
        csv_dir: Directory the CSV files reside in.
        drop_tables: Existing tables will be dropped if True.
        safe: Load can be faster if set to False, but can corrupt the db
            if it crashes / interrupted.
        recursive: Load all dependent tables if set to True.

    Returns:
        Nothing.

    """
    # Use all tables if no table is provided.
    models = models or tables.MODELS
    logger.debug("Tables to be loaded: {tables}", tables=models)

    # Load tables faster.
    if not safe:
        database.synchronous = 0
        database.journal_mode = "memory"
        logger.debug(
            "Safe option turned off: synchronous: {sync}, journal-mode: {mode}",
            sync=database.synchronous,
            mode=database.journal_mode,
        )

    logger.debug("Opening database {uri}", uri=database.database)
    with database.atomic():
        logger.debug("Opened database {uri}", uri=database.database)
        # Enable foreign keys
        database.foreign_keys = 1
        logger.debug("Foreign key set to {fk}", fk=database.foreign_keys)

        # Bind the database.
        database.bind(models, bind_refs=recursive, bind_backrefs=recursive)
        logger.debug("Bound database.")

        # Drop tables if asked.
        if drop_tables:
            database.drop_tables(models)
            logger.debug("Dropped tables: {tables}", tables=models)

        # Create tables.
        database.create_tables(models)
        logger.debug("Tables created.")

        # Run through the CSV files and load the data.
        for model in models:
            try:
                # Read CSV data.
                csv_file_path = Path(csv_dir) / f"{model._meta.table_name}.csv"
                csv_dict_reader = csv.DictReader(csv_file_path.open(mode="r"))

                # Insert CSV data to table in batches.
                # http://docs.peewee-orm.com/en/latest/peewee/querying.html#inserting-rows-in-batches
                for batch in peewee.chunked(csv_dict_reader, 100):
                    model.insert_many(batch).execute()

                logger.debug("Written table {table} into database.", table=model._meta.table_name)

            except IOError:
                # Log the error and continue the next loop.
                logger.error("CSV file not found: {csv_file}", csv_file=csv_file_path.name)
                continue

    return True



"""Provides the pseudo-random number generator used in various places."""
from typing import Generator, List, Tuple, TypeVar

import attr

P = TypeVar("P", bound="PRNG")


@attr.s(slots=True, auto_attribs=True, cmp=False)
class PRNG:
    """A linear congruential random number generator.

    Usage:
        >>> prng = PRNG()
        >>> prng()
        0
        >>> prng()
        59774

    References:
        https://bulbapedia.bulbagarden.net/wiki/Pseudorandom_number_generation_in_Pokémon
        https://www.smogon.com/ingame/rng/pid_iv_creation#pokemon_random_number_generator
    """

    seed: int = attr.ib(validator=attr.validators.instance_of(int), default=0)
    _gen: int = attr.ib(validator=attr.validators.in_(range(1, 8)), default=3)
    _initial_seed: int = attr.ib(init=False)

    def __attrs_post_init__(self: P) -> None:
        """Record the initial seed right after instantiation."""
        self._initial_seed = self.seed

    def _generator(self: P) -> Generator:
        if self._gen == 3:
            while True:
                self.seed = (0x41C64E6D * self.seed + 0x6073) & 0xFFFFFFFF
                yield self.seed >> 16
        else:
            raise ValueError(f"Gen. {self._gen} PRNG is not supported yet.")

    def __call__(self: P) -> int:
        """Move to the next random number."""
        try:
            return next(self._generator())
        except StopIteration:
            self.reset()
            return next(self._generator())

    def reset(self: P) -> None:
        """Reset the generator with the initial seed."""
        self.seed = self._initial_seed

    def next_(self: P, n: int) -> List[int]:
        """Generate the next n random numbers."""
        return [self() for _ in range(n)]

    def generate_pid_and_iv(self: P, method: int = 2) -> Tuple[int, int]:
        """Generate the PID and IVs using the internal generator.

        The generated 32-bit IVs is different from how it is actually
        stored.

        Method 1: [PID] [PID] [IVs] [IVs]
        Method 2: [PID] [PID] [xxxx] [IVs] [IVs]
        Method 4: [PID] [PID] [IVs] [xxxx] [IVs]

        Methods 2 and 4 are only used in Pokemon Ruby, Sapphire, Emerald,
        Fire Red and Leaf Green (RSEFRLG) to produce wild Pokemon.

        All the Pokemon you catch in these games that are not wild Pokemon
        are created using Method 1.

        The criterion for choosing whether to use Method 1, 2, or 4 in the
        creation of wild Pokemon in RSEFRLG seems to be arbitrary.

        Checkout [this link](https://www.smogon.com/ingame/rng/pid_iv_creation#rng_pokemon_generation)  # noqa: E501
        for more information on Method 1, 2, and 4.

        Args:
            method: 1, 2, or 4.

        Raises:
            ValueError: if the method is not in (1, 2, 4).

        Returns:
            a tuple of two integers, in the order of 'PID' and 'IVs'.
        """
        if method not in (1, 2, 4):
            raise ValueError(
                "Only methods 1, 2, 4 are supported. For more information on "
                "the meaning of the methods, see "
                "https://www.smogon.com/ingame/rng/pid_iv_creation#rng_pokemon_generation"
                " for help."
            )

        return self._generate_pid(), self._generate_iv(method)

    def _generate_pid(self: P) -> int:
        """Create the Personality ID (PID) for a Pokémon.

        Returns:
            int
        """
        pid_src_1, pid_src_2 = self.next_(2)
        return pid_src_1 + (pid_src_2 << 16)

    def _generate_iv(self: P, method: int = 2) -> int:
        """Create the number used to generate a Pokémon's IVs.

        Args:
            method: the Pokémon generation method. Valid values are 1, 2, and 4.

        Returns:
            int
        """
        if method == 1:
            iv_src_1, iv_src_2 = self.next_(2)
        elif method == 2:
            _, iv_src_1, iv_src_2 = self.next_(3)
        else:  # method == 4:
            iv_src_1, _, iv_src_2 = self.next_(3)

        return iv_src_1 + (iv_src_2 << 16)

    def random(self: P) -> float:
        """Generate a random number from the uniform distribution [0, 1).

        Returns:
            A random number between 0 and 1.
        """
        return self() / 0x10000



"""Console script for pokemaster2."""

import click
from loguru import logger

from pokemaster2 import __version__
from pokemaster2.db import io


@click.group()
@click.version_option(version=__version__)
def main() -> int:
    """Console script for pokemaster2."""
    # click.echo("Replace this message by putting your code into pokemaster2.cli.main")
    # click.echo("See click documentation at https://click.palletsprojects.com/")
    return 0


@main.command("load")
@click.option("-C", "--csv-dir", default=None)
@click.option("-U", "--uri", default=None)
@click.option("-D", "--drop-tables", type=bool, default=True)
@click.option("-S", "--safe", type=bool, default=True)
@click.option("-R", "--recursive", type=bool, default=True)
def cli_load(csv_dir: str, uri: str, drop_tables: bool, safe: bool, recursive: bool) -> None:
    """Load Pokédex data into a database from CSV files."""
    logger.info("Running command `load`.")
    io.load(
        database=io.get_database(uri),
        csv_dir=io.get_csv_dir(csv_dir),
        models=None,
        drop_tables=drop_tables,
        safe=safe,
        recursive=recursive,
    )
    logger.debug("Successfully loaded database.")
    return 0


if __name__ == "__main__":
    main()  # pragma: no cover
